import logging
from pathlib import Path

from magnus import defaults
from magnus.catalog import FileSystemCatalog

logger = logging.getLogger(defaults.NAME)


class K8sPVCatalog(FileSystemCatalog):

    service_name = 'k8s-pvc'

    class Config(FileSystemCatalog.Config):
        persistent_volume_name: str
        mount_path: str

    @property
    def catalog_location(self) -> str:
        """
        Get the catalog location from the config.
        If its not defined, use the magnus default

        Returns:
            str: The catalog location as defined by the config or magnus default '.catalog'
        """
        return str(Path(self.config.mount_path) / self.config.catalog_location)  # type: ignore



import json
import logging
import tempfile

import botocore
from magnus import defaults, exceptions
from magnus.datastore import BaseRunLogStore, RunLog

logger = logging.getLogger(defaults.NAME)


try:
    from magnus_extensions.aws import AWSConfigMixin
except ImportError as _e:  # pragma: no cover
    msg = (
        "AWS Dependencies are not installed!!"
    )
    raise Exception(msg) from _e


class S3Store(BaseRunLogStore, AWSConfigMixin):
    """
    S3 implementation of Run Log store

    Example config:
    run_log_store:
      type: s3
      config:
        s3_bucket: The S3 bucket to use
        prefix: Any bucket prefix that you want to attach

    """
    service_name = 's3'

    class Config(AWSConfigMixin.Config):
        s3_bucket: str
        prefix: str = ''

    @property
    def s3_bucket(self) -> str:
        """
        Returns the S3 bucket name from the config

        Returns:
            str: The name of the s3 bucket as per the config
        """
        return self.config.s3_bucket

    @property
    def prefix(self):
        """
        Return the prefix if present in the config. Or an empty string
        """
        return self.config.prefix.strip('/')

    def write_to_bucket(self, run_log: RunLog):
        """Writes the run log to S3 bucket

        Args:
            run_log (RunLog): The run log to write to bucket

        Raises:
            Exception: When the upload fails
        """
        boto3_session = self.get_boto3_session()
        s3_client = boto3_session.client('s3')
        temp_file = tempfile.NamedTemporaryFile()

        try:
            with open(temp_file.name, 'w', encoding='utf-8') as fw:
                json.dump(run_log.dict(), fw, ensure_ascii=True, indent=4)  # pylint: disable=no-member

            if self.prefix:
                object_key = f'{self.prefix}/{run_log.run_id}.json'
            else:
                object_key = f'{run_log.run_id}.json'

            s3_client.upload_file(temp_file.name, self.s3_bucket, object_key)
        except botocore.exceptions.ClientError as _e:
            if _e.response['Error']['Code'] == "404":
                logger.exception(f'Upload failed to {self.s3_bucket}')
            raise Exception(_e) from _e
        finally:
            temp_file.close()

    def get_from_bucket(self, run_id: str) -> RunLog:
        """
        Get a Run log object from the S3 bucket for the run id.

        Args:
            run_id (str): The run id for which we want the run log

        Raises:
            exceptions.RunLogNotFoundError: If the run log by the run id is not found
            Exception: If the access to bucket is not allowed

        Returns:
            RunLog: The run log of the run id
        """
        boto3_session = self.get_boto3_session()
        s3_client = boto3_session.client('s3')
        temp_file = tempfile.NamedTemporaryFile()

        try:
            if self.prefix:
                key = f'{self.prefix}/{run_id}.json'
            else:
                key = f'{run_id}.json'

            logger.info(f'Trying to download {key} from {self.s3_bucket}')
            s3_client.download_file(self.s3_bucket, key, temp_file.name)

            json_str = json.load(open(temp_file.name, 'rb'))
            run_log = RunLog(**json_str)
            return run_log
        except botocore.exceptions.ClientError as _e:
            if _e.response['Error']['Code'] == "404":
                logger.exception(f'{key} does not exist in {self.s3_bucket}')
                raise exceptions.RunLogNotFoundError(run_id)
            raise
        finally:
            temp_file.close()

    def create_run_log(self, run_id: str, dag_hash: str = '', use_cached: bool = False,
                       tag: str = '', original_run_id: str = '', status: str = defaults.CREATED, **kwargs) -> RunLog:
        # Creates a Run log
        # Adds it to the db
        try:
            self.get_run_log_by_id(run_id=run_id, full=False)
            raise exceptions.RunLogExistsError(run_id=run_id)
        except exceptions.RunLogNotFoundError:
            pass

        logger.info(f'{self.service_name} Creating a Run Log for : {run_id}')
        run_log = RunLog(run_id=run_id, dag_hash=dag_hash, use_cached=use_cached,
                         tag=tag, original_run_id=original_run_id, status=status)
        self.write_to_bucket(run_log)
        return run_log

    def get_run_log_by_id(self, run_id, full=True, **kwargs):
        # Returns the run_log defined by id
        # Raises Exception if not found
        try:
            logger.info(f'{self.service_name} Getting a Run Log for : {run_id}')
            run_log = self.get_from_bucket(run_id)
            return run_log
        except Exception as e:
            raise exceptions.RunLogNotFoundError(run_id) from e

    def put_run_log(self, run_log, **kwargs):
        # Puts the run_log into the database
        logger.info(f'{self.service_name} Putting the run log in the DB: {run_log.run_id}')
        self.write_to_bucket(run_log)



from pathlib import Path

from magnus.integration import BaseIntegration


class LocalContainerCompute(BaseIntegration):
    """
    Integration pattern between Local container and S3  catalog
    """
    executor_type = 'local-container'
    service_type = 'catalog'  # One of secret, catalog, datastore
    service_provider = 's3'  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        write_to = self.service.get_aws_credentials_file()
        self.executor.volumes[str(Path(write_to).resolve())] = {
            'bind': '/root/.aws/credentials',
            'mode': 'ro'
        }



import logging
import shlex
from typing import Any, List, Optional, Union

from magnus import defaults, integration, utils
from magnus.executor import BaseExecutor
from magnus.graph import Graph
from magnus.nodes import BaseNode
from pydantic import BaseModel
from ruamel.yaml import YAML

logger = logging.getLogger(defaults.NAME)


class SecretEnvVar(BaseModel):
    environment_variable: str
    secret_name: str
    secret_key: str

    def dict(self, *args, **kwargs):
        return {
            "name": self.environment_variable,
            "valueFrom": {
                "secretKeyRef": {
                    "name": self.secret_name,
                    "key": self.secret_key
                }
            }
        }


class EnvVar(BaseModel):
    name: str
    value: Any

    def dict(self, *args, **kwargs):
        return_value = {
            "name": self.name
        }

        if self.value:
            return_value["value"] = self.value

        return return_value


class Request(BaseModel):
    memory: str
    cpu: str


class Limit(BaseModel):
    memory: str
    cpu: str
    gpu: int = 0

    def dict(self, *args, **kwargs) -> dict:
        resource = {
            "cpu": self.cpu,
            "memory": self.memory
        }
        if self.gpu:
            # TODO: This should be via config to allow users to specify the vendor
            resource["nvidia.com/gpu"] = self.gpu

        return resource


class VolumeMount(BaseModel):
    mountPath: str
    name: str


class Container(BaseModel):
    command: List[str]
    image: str
    limits: Limit
    requests: Request
    imagePullPolicy: str = "IfNotPresent"
    retry: int = 1
    env: List[Union[SecretEnvVar, EnvVar]] = []
    volumeMounts: List[VolumeMount] = []

    def dict(self, *args, **kwargs) -> dict:
        return {
            "command": self.command,
            "image": self.image,
            "imagePullPolicy": self.imagePullPolicy,
            "resources": {
                "limits": self.limits.dict(),
                "requests": self.requests.dict()
            },
            "retryStrategy": {"limit": str(self.retry)},
            "env": [e.dict() for e in self.env],
            "volumeMounts": [v.dict() for v in self.volumeMounts]
        }


class TaskTemplate(BaseModel):
    name: str
    template: str
    dependencies: List[str] = []


class ContainerTemplate(BaseModel):
    name: str
    container: Container


class DagTemplate(BaseModel):
    name: str = "magnus-dag"
    tasks: List[TaskTemplate] = []

    def dict(self, *args, **kwargs):
        return {
            "name": self.name,
            "dag": {
                "tasks": [task.dict() for task in self.tasks]
            }
        }


class Volume(BaseModel):
    name: str
    claim: str

    def dict(self, *args, **kwargs) -> dict:
        return {
            "name": self.name,
            "persistentVolumeClaim": {
                "claimName": self.claim
            }
        }


class Spec(BaseModel):
    entrypoint: str = "magnus-dag"
    templates: Union[DagTemplate, ContainerTemplate] = []
    serviceAccountName: str = "pipeline-runner"
    arguments: List[EnvVar] = []
    volumes: List[Volume] = []

    def dict(self, *args, **kwargs):
        return {
            "entrypoint": self.entrypoint,
            "volumes": [v.dict() for v in self.volumes],
            "templates": [template.dict() for template in self.templates],
            "serviceAccountName": self.serviceAccountName,
            "arguments": {
                "parameters": [env.dict() for env in self.arguments]
            }
        }


class WorkSpec(BaseModel):
    apiVersion: str = "argoproj.io/v1alpha1"
    kind: str = "Workflow"
    metadata: dict = {"generateName": "magnus-dag-"}
    spec: Optional[Spec]


class ArgoExecutor(BaseExecutor):

    service_name = "argo"
    run_id_placeholder = "{{workflow.parameters.run_id}}"

    class Config(BaseExecutor.Config):
        docker_image: str
        output_file: str = 'pipeline.yaml'
        cpu_limit: str = "250m"
        memory_limit: str = "1G"
        cpu_request: str = ""
        memory_request: str = ""
        gpu_limit: int = 0
        enable_caching: bool = False
        image_pull_policy: str = "Always"
        secrets_from_k8s: dict = {}  # EnvVar=SecretName:Key
        persistent_volumes: dict = {}  # volume-name:mount_path

    def __init__(self, config: dict = None):
        super().__init__(config)
        self.persistent_volumes = {}

        for i, (volume_name, mount_path) in enumerate(self.config.persistent_volumes.items()):
            self.persistent_volumes[f"executor-{i}"] = (volume_name, mount_path)

    def prepare_for_graph_execution(self):
        """
        This method would be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we need not do anything.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_traversal(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_traversal(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_traversal(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_traversal(self, self.experiment_tracker)

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_execution(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_execution(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_execution(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_execution(self, self.experiment_tracker)

        self._set_up_run_log(exists_ok=True)

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self.run_log_store.add_step_log(step_log, self.run_id)

        super()._execute_node(node, map_variable=map_variable, **kwargs)

        # Implicit fail
        if self.dag:
            # functions and notebooks do not have dags
            _, next_node_name = self._get_status_and_next_node_name(node, self.dag, map_variable=map_variable)
            if next_node_name:
                # Terminal nodes do not have next node name
                next_node = self.dag.get_node_by_name(next_node_name)

                if next_node.node_type == defaults.FAIL:
                    self.execute_node(next_node, map_variable=map_variable)

        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
        if step_log.status == defaults.FAIL:
            raise Exception(f'Step {node.name} failed')

    def get_parameters(self):
        parameters = utils.get_user_set_parameters()
        if self.parameters_file:
            parameters.update(utils.load_yaml(self.parameters_file))
        return parameters

    def get_clean_name(self, node_name: str):
        return node_name.replace(" ", "-")

    def create_container_template(self, working_on: BaseNode, command: str, add_parameters: bool = False):
        command = shlex.split(command)
        mode_config = self._resolve_node_config(working_on)
        secrets = mode_config.get("secrets_from_k8s", {})

        docker_image = mode_config['docker_image']
        cpu_limit = mode_config.get('cpu_limit', self.config.cpu_limit)
        memory_limit = mode_config.get('memory_limit', self.config.memory_limit)

        cpu_request = mode_config.get('cpu_request', self.config.cpu_request) or cpu_limit
        memory_request = mode_config.get('memory_request', self.config.memory_request) or memory_limit

        gpu_limit = mode_config.get('gpu_limit', self.config.gpu_limit)

        request = Request(memory=memory_request, cpu=cpu_request)
        limits = Limit(memory=memory_limit, cpu=cpu_limit, gpu=gpu_limit)

        image_pull_policy = mode_config.get("image_pull_policy", self.config.image_pull_policy)

        container = Container(command=command, image=docker_image, limits=limits, requests=request,
                              imagePullPolicy=image_pull_policy, retry=working_on._get_max_attempts())
        for secret_env, k8_secret in secrets.items():
            try:
                secret_name, key = k8_secret.split(':')
            except Exception as _e:
                msg = (
                    "K8's secret should be of format EnvVar=SecretName:Key"
                )
                raise Exception(msg) from _e
            secret = SecretEnvVar(environment_variable=secret_env, secret_name=secret_name, secret_key=key)
            container.env.append(secret)

        if add_parameters or working_on.name == self.dag.start_at:  # short circuit when forcing to add parameters
            for key, _ in self.get_parameters().items():
                # Get the value from work flow parameters for dynamic behavior
                env_var = EnvVar(name=defaults.PARAMETER_PREFIX + key, value="{{workflow.parameters." + key + "}}")
                container.env.append(env_var)

        visited_claims = {}
        for volume_name, claim in self.persistent_volumes.items():
            claim_name, mount_path = claim

            # If the volume is already mounted, we cannot mount it again.
            if claim_name in visited_claims:
                msg = (
                    "The same persistent volume claim has already been used in the pipeline by another service"
                )
                raise Exception(msg)
            visited_claims[claim_name] = claim_name

            container.volumeMounts.append(VolumeMount(name=volume_name, mountPath=mount_path))

        container_template = ContainerTemplate(name=self.get_clean_name(working_on.name), container=container)

        return container_template

    def get_templates(self, dag):
        current_node = dag.start_at
        previous_node = None

        container_templates = {}
        task_templates = {}
        while True:
            working_on = dag.get_node_by_name(current_node)
            if working_on.is_composite:
                raise NotImplementedError('Composite nodes are not yet implemented')

            if previous_node == current_node:
                raise Exception('Potentially running in a infinite loop')

            command = utils.get_node_execution_command(self, working_on,
                                                       over_write_run_id=self.run_id_placeholder)
            container_template = self.create_container_template(working_on=working_on, command=command)

            if current_node not in container_templates:
                container_templates[current_node] = container_template

            clean_name = self.get_clean_name(working_on.name)
            task_template = TaskTemplate(name=clean_name, template=clean_name)
            if previous_node:
                task_template.dependencies.append(self.get_clean_name(previous_node))

            task_templates[current_node] = task_template

            previous_node = current_node

            if working_on.node_type in ['success', 'fail']:
                break

            current_node = working_on._get_next_node()

        return [template for k, template in container_templates.items()], \
            [template for k, template in task_templates.items()]

    def add_volumes_to_specification(self, specification: Spec):
        visited_claims = {}
        for volume_name, claim in self.persistent_volumes.items():
            claim_name, mount_path = claim

            # If the volume is already mounted, we cannot mount it again.
            if claim_name in visited_claims:
                msg = (
                    "The same persistent volume claim has already been used in the pipeline by another service"
                )
                raise Exception(msg)
            visited_claims[claim_name] = claim_name
            specification.volumes.append(Volume(name=volume_name, claim=claim_name))

    def execute_graph(self, dag: Graph, map_variable: dict = None, **kwargs):
        workspec = WorkSpec()
        specification = Spec()
        workspec.spec = specification

        for key, value in self.get_parameters().items():
            # Get the value from work flow parameters for dynamic behavior
            env_var = EnvVar(name=key, value=value)
            specification.arguments.append(env_var)

        run_id_var = EnvVar(name="run_id", value="{{workflow.uid}}")
        specification.arguments.append(run_id_var)

        # Add volumes to specification
        self.add_volumes_to_specification(specification)

        container_templates, task_templates = self.get_templates(dag=dag)
        specification.templates.extend(container_templates)
        dag_template = DagTemplate(tasks=task_templates)
        specification.templates.extend([dag_template])
        yaml = YAML()
        with open(self.config.output_file, 'w') as f:
            yaml.dump(workspec.dict(), f)

    def execute_job(self, node: BaseNode):
        workspec = WorkSpec()
        specification = Spec()
        workspec.spec = specification

        for key, value in self.get_parameters().items():
            # Get the value from work flow parameters for dynamic behavior
            env_var = EnvVar(name=key, value=value)
            specification.arguments.append(env_var)

        run_id_var = EnvVar(name="run_id", value="{{workflow.uid}}")
        specification.arguments.append(run_id_var)

        # Add volumes to specification
        self.add_volumes_to_specification(specification)

        command = utils.get_job_execution_command(self, node, over_write_run_id=self.run_id_placeholder)
        container_template = self.create_container_template(working_on=node, command=command, add_parameters=True)

        clean_name = self.get_clean_name(node.name)
        task_template = TaskTemplate(name=clean_name, template=clean_name)

        specification.templates.extend([container_template])
        dag_template = DagTemplate(tasks=[task_template])
        specification.templates.extend([dag_template])
        yaml = YAML()
        with open(self.config.output_file, 'w') as f:
            yaml.dump(workspec.dict(), f)

    def send_return_code(self, stage='traversal'):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        if stage != 'traversal':  # traversal does no actual execution, so return code is pointless
            run_id = self.run_id

            run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
            if run_log.status == defaults.FAIL:
                raise Exception('Pipeline execution failed')



import fnmatch
import logging
import os
from pathlib import Path

from magnus import defaults, utils
from magnus.catalog import (BaseCatalog, get_run_log_store,
                            is_catalog_out_of_sync)

logger = logging.getLogger(defaults.NAME)

try:
    from magnus_extensions.aws import AWSConfigMixin
except ImportError as _e:  # pragma: no cover
    msg = (
        "AWS Dependencies are not installed!!"
    )
    raise Exception(msg) from _e


class S3Catalog(BaseCatalog, AWSConfigMixin):
    """
    TODO: The catalog relative paths might be wrong and sync between runs might not work

    A S3 based catalog.

    Example config:
    catalog:
      type: s3
      config:
        compute_data_folder : data/
        s3_bucket: Bucket name
        prefix: Any prefix in S3 to search for the catalog files
    """
    service_name = 's3'

    class Config(BaseCatalog.Config, AWSConfigMixin.Config):
        s3_bucket: str
        prefix: str = ''

    @property
    def prefix(self) -> str:
        """
        Return the prefix if present in the config. Or an empty string
        """
        return self.config.prefix.strip('/')

    def get_s3(self):
        """Gets a s3 client object from a boto3 session.

        Returns:
            botocore.client: s3 client object
        """
        boto_session = self.get_boto3_session()
        return boto_session.client('s3')

    @property
    def bucket_name(self) -> str:
        """Gets the bucket name.

        Returns:
            str: Name of the S3 bucket for the catalog.
        """
        return self.config.s3_bucket

    def check_s3_access(self, s3_client, bucket):
        """Checks access to a given S3 bucket

        Args:
            s3_client (botocore.client): S3 client
            bucket (str): Bucket name

        Raises:
            Exception: If can't access the bucket then an Exception is raised
        """
        check_access = s3_client.head_bucket(Bucket=bucket)

        if check_access['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Exception(f'Expected Catalog s3 Bucket {bucket} does not exist, or you do not have access')

    def get(self, name: str, run_id: str, compute_data_folder: str = None, **kwargs):
        """Gets files from S3 store and moves to compute data folder

        Args:
            name (str): The glob pattern
            run_id (str): The ID of the current run
            compute_data_folder (str, optional): The path of the compute data to sync to. Defaults to None.

        Returns:
            list: List of data catalogs
        """
        copy_to = self.compute_data_folder

        if compute_data_folder:
            copy_to = compute_data_folder

        s3_client = self.get_s3()
        self.check_s3_access(s3_client, self.bucket_name)

        s3_prefix = f'{run_id}'
        if self.prefix:
            s3_prefix = f'{self.prefix}/{s3_prefix}'

        if copy_to != '.':
            s3_prefix = f'{s3_prefix}/{copy_to}'

        page_iter = s3_client.get_paginator('list_objects_v2').paginate(
            Bucket=self.bucket_name, Prefix=s3_prefix)

        logger.debug('Contents from S3')
        for page in page_iter:
            logger.debug(page)

        search_name = f'{s3_prefix}/{name}'
        if name == '*':
            search_name = '*'

        s3_files = []
        for page in page_iter:
            try:
                s3_files += fnmatch.filter([file['Key'] for file in page['Contents']], search_name)
            except KeyError:
                logger.warning("Did not find any objects matching the catalog pattern")
                return

        data_catalogs = []
        run_log_store = get_run_log_store()

        for file in s3_files:
            file_obj = s3_client.get_object(Bucket=self.bucket_name, Key=file)

            # Remove run_id and s3 prefix as they are specific to cataloging method
            write_path = Path(file.replace(f'{run_id}/', '').replace(self.prefix, "").lstrip(os.sep))
            with write_path.open('wb') as f:
                f.write(file_obj['Body'].read())

            data_catalog = run_log_store.create_data_catalog(str(write_path))
            data_catalog.catalog_handler_location = f"s3://{self.bucket_name}"
            data_catalog.catalog_relative_path = str(file)
            data_catalog.data_hash = utils.get_data_hash(write_path)
            data_catalog.stage = 'get'
            data_catalogs.append(data_catalog)

        return data_catalogs

    def put(self, name: str, run_id: str, compute_data_folder: str = None, synced_catalogs=None, **kwargs):
        """Moves the data from the compute data folder to the s3 data catalog

        Args:
            name (str): The glob string pattern
            run_id (str): [description]
            compute_data_folder ([type], optional): [description]. Defaults to None.
            synced_catalogs ([type], optional): [description]. Defaults to None.

        Raises:
            Exception: [description]

        Returns:
            [type]: [description]
        """

        copy_from = self.compute_data_folder
        if compute_data_folder:
            copy_from = compute_data_folder

        if not utils.does_dir_exist(copy_from):
            raise Exception(f'Expected compute data folder to be present at: {copy_from} but not found')

        s3_client = self.get_s3()
        self.check_s3_access(s3_client, bucket=self.bucket_name)

        glob_files = Path(copy_from).glob(name)

        data_catalogs = []
        run_log_store = get_run_log_store()

        for file in glob_files:
            if file.is_dir():
                # Need not add a data catalog for the folder
                continue
            relative_file_path = file.relative_to('.')

            file_key = f'{run_id}/{relative_file_path}'
            if self.prefix:
                file_key = f'{self.prefix}/{file_key}'

            data_catalog = run_log_store.create_data_catalog(str(relative_file_path))
            data_catalog.catalog_handler_location = f"s://{self.bucket_name}"
            data_catalog.catalog_relative_path = file_key
            data_catalog.data_hash = utils.get_data_hash(str(file))
            data_catalog.stage = 'put'
            data_catalogs.append(data_catalog)

            if is_catalog_out_of_sync(data_catalog, synced_catalogs):
                logger.info(f'{data_catalog.name} was found to be changed, syncing')

                s3_client.upload_file(Filename=str(file.resolve()), Bucket=self.bucket_name,
                                      Key=file_key)
            else:
                logger.info(f'{data_catalog.name} was found to be unchanged, ignoring syncing')
        return data_catalogs

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Given data catalogs from a previous run, sync them to the catalog of the run given by run_id

        Args:
            previous_run_id (str): The run_id of the previous run
            run_id (str): The run_id to which the data catalogs should be synced to.

        """
        s3_client = self.get_s3()
        self.check_s3_access(s3_client, self.bucket_name)

        s3_files = []
        look_at = previous_run_id
        if self.prefix:
            look_at = self.prefix + '/' + look_at

        page_iter = s3_client.get_paginator('list_objects_v2').paginate(Bucket=self.bucket_name,
                                                                        Prefix=look_at)
        for page in page_iter:
            s3_files += [file['Key'] for file in page['Contents']]

        for file in s3_files:
            s3_client.copy_object(Bucket=self.bucket_name,
                                  CopySource=file,
                                  Key=file.replace(previous_run_id, run_id))



import logging
from pathlib import Path

from magnus import defaults
from magnus.datastore import ChunkedFileSystemRunLogStore

logger = logging.getLogger(defaults.NAME)


class ChunkedK8PersistentVolumeRunLogstore(ChunkedFileSystemRunLogStore):
    """
    Uses the K8s Persistent Volumes to store run logs.
    """

    service_name = "chunked-k8s-pvc"

    class Config(ChunkedFileSystemRunLogStore.Config):
        persistent_volume_name: str
        mount_path: str

    @property
    def log_folder_name(self) -> str:
        return str(Path(self.config.mount_path) / self.config.log_folder)



import logging
import shlex

from kfp.onprem import mount_pvc
from magnus import defaults, integration, utils
from magnus.executor import BaseExecutor
from magnus.graph import Graph
from magnus.nodes import BaseNode
from ruamel.yaml import YAML

logger = logging.getLogger(defaults.NAME)

try:
    from kfp import dsl
    from kfp.compiler.compiler import Compiler
    from kubernetes.client import V1EnvVar, V1EnvVarSource, V1SecretKeySelector
except ImportError as _e:
    msg = (
        "Kubeflow Dependencies have not been installed!!"
    )
    raise Exception(msg) from _e


_EXECUTOR = None
_GRAPH = None
_PARAMETERS = None
_NODE = None


def secret_to_env_var(secret_name: str, secret_key: str, env_var_name: str) -> V1EnvVar:
    """Turns a K8s secret into an environment variable for use in a kubeflow component"""
    return V1EnvVar(
        name=env_var_name,
        value_from=V1EnvVarSource(secret_key_ref=V1SecretKeySelector(name=secret_name, key=secret_key)),
    )


def disable_cache(task):
    """Disables caching on passed task"""
    task.execution_options.caching_strategy.max_cache_staleness = "P0D"


class KubeFlowExecutor(BaseExecutor):

    service_name = "kfp"
    run_id_placeholder = dsl.RUN_ID_PLACEHOLDER
    replace_run_id = "{{workflow.parameters.run_id}}"

    class Config(BaseExecutor.Config):
        docker_image: str
        output_file: str = 'pipeline.yaml'
        cpu_limit: str = "250m"
        memory_limit: str = "1G"
        gpu_limit: int = 0
        cpu_request: str = ""
        memory_request: str = ""
        enable_caching: bool = False
        image_pull_policy: str = "Always"
        secrets_from_k8s: dict = {}  # EnvVar=SecretName:Key
        persistent_volumes: dict = {}  # volume-name:mount_path

    def __init__(self, config: dict = None):
        super().__init__(config)
        self.persistent_volumes = {}

        for i, (volume_name, mount_path) in enumerate(self.config.persistent_volumes.items()):
            self.persistent_volumes[f"executor-{i}"] = (volume_name, mount_path)

    def prepare_for_graph_execution(self):
        """
        This method would be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we need not do anything.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_traversal(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_traversal(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_traversal(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_traversal(self, self.experiment_tracker)

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_execution(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_execution(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_execution(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_execution(self, self.experiment_tracker)

        self._set_up_run_log(exists_ok=True)

    def execute_job(self, node: BaseNode):
        global _EXECUTOR, _PARAMETERS, _NODE

        _EXECUTOR = self
        _NODE = node

        compiler = Compiler()
        # set parameters as Kfp parameters
        parameters = utils.get_user_set_parameters()
        if self.parameters_file:
            parameters.update(utils.load_yaml(self.parameters_file))
        _PARAMETERS = parameters

        pipeline_params = []
        for param_key, param_value in parameters.items():
            param = dsl.PipelineParam(name=param_key, value=param_value)
            pipeline_params.append(param)

        pipeline_params.append(dsl.PipelineParam(name='run_id', value=self.run_id_placeholder))

        pipeline = compiler.create_workflow(pipeline_func=convert_to_kfp_job,
                                            pipeline_name='magnus_dag', params_list=pipeline_params)

        # compiler.compile(convert_to_kfp, self.config.output_file)
        yaml = YAML()
        with open(self.config.output_file, 'w') as f:
            yaml.dump(pipeline, f)

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self.run_log_store.add_step_log(step_log, self.run_id)

        super()._execute_node(node, map_variable=map_variable, **kwargs)

        # Implicit fail
        if self.dag:
            # notebook or func does not have dag definitions.
            _, next_node_name = self._get_status_and_next_node_name(node, self.dag, map_variable=map_variable)
            # Terminal nodes do not have next node
            if next_node_name:
                next_node = self.dag.get_node_by_name(next_node_name)

                if next_node.node_type == defaults.FAIL:
                    self.execute_node(next_node, map_variable=map_variable)

        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
        if step_log.status == defaults.FAIL:
            raise Exception(f'Step {node.name} failed')

    def create_container_op(self, working_on: BaseNode, command: str):
        global _GRAPH
        command = shlex.split(command)
        mode_config = self._resolve_node_config(working_on)

        docker_image = mode_config['docker_image']
        secrets = mode_config.get("secrets_from_k8s", {})

        operator = dsl.ContainerOp(name=working_on._command_friendly_name(), image=docker_image, command=command)

        cpu_limit = mode_config.get('cpu_limit', self.config.cpu_limit)
        memory_limit = mode_config.get('memory_limit', self.config.memory_limit)

        cpu_request = mode_config.get('cpu_request', self.config.cpu_request) or cpu_limit
        memory_request = mode_config.get('memory_request', self.config.memory_request) or memory_limit

        gpu_limit = mode_config.get("gpu_limit", self.config.gpu_limit)

        operator.set_memory_limit(memory_limit).set_memory_request(
            memory_request).set_cpu_request(cpu_request).set_cpu_limit(cpu_limit)

        if gpu_limit:
            operator.set_gpu_limit(gpu_limit)

        operator.set_retry(str(working_on._get_max_attempts()))

        image_pull_policy = mode_config.get("image_pull_policy", self.config.image_pull_policy)
        operator.container.set_image_pull_policy(image_pull_policy)

        for secret_env, k8_secret in secrets.items():
            try:
                secret_name, key = k8_secret.split(':')
            except Exception as _e:
                msg = (
                    "K8's secret should be of format EnvVar=SecretName:Key"
                )
                raise Exception(msg) from _e
            operator.add_env_variable(secret_to_env_var(secret_name=secret_name,
                                      secret_key=key, env_var_name=secret_env))

        visited_claims = {}
        for volume_name, claim in self.persistent_volumes.items():
            claim_name, mount_path = claim

            # If the volume is already mounted, we cannot mount it again.
            if claim_name in visited_claims:
                msg = (
                    "The same persistent volume claim has already been used in the pipeline by another service"
                )
                raise Exception(msg)
            visited_claims[claim_name] = claim_name

            operator.apply(mount_pvc(pvc_name=claim_name, volume_name=volume_name, volume_mount_path=mount_path))

        if _GRAPH and _GRAPH.start_at == working_on.name:
            global _PARAMETERS
            parameters = _PARAMETERS

            for key, _ in parameters.items():
                operator.add_env_variable(V1EnvVar(name=defaults.PARAMETER_PREFIX + key,
                                          value="{{workflow.parameters." + key + "}}"))

        return operator

    def execute_graph(self, dag: Graph, map_variable: dict = None, **kwargs):
        global _EXECUTOR, _GRAPH, _PARAMETERS

        _EXECUTOR = self
        _GRAPH = dag

        compiler = Compiler()
        # set parameters as Kfp parameters
        parameters = utils.get_user_set_parameters()
        if self.parameters_file:
            parameters.update(utils.load_yaml(self.parameters_file))
        _PARAMETERS = parameters

        pipeline_params = []
        for param_key, param_value in parameters.items():
            param = dsl.PipelineParam(name=param_key, value=param_value)
            pipeline_params.append(param)

        pipeline_params.append(dsl.PipelineParam(name='run_id', value=self.run_id_placeholder))

        pipeline = compiler.create_workflow(pipeline_func=convert_to_kfp,
                                            pipeline_name='magnus_dag', params_list=pipeline_params)
        # compiler.compile(convert_to_kfp, self.config.output_file)
        yaml = YAML()
        with open(self.config.output_file, 'w') as f:
            yaml.dump(pipeline, f)

    def send_return_code(self, stage='traversal'):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        if stage != 'traversal':  # traversal does no actual execution, so return code is pointless
            run_id = self.run_id

            run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
            if run_log.status == defaults.FAIL:
                raise Exception('Pipeline execution failed')


def convert_to_kfp_job():
    """
    The function that makes the workflow
    """
    self = _EXECUTOR
    node = _NODE
    parameters = _PARAMETERS

    command = utils.get_job_execution_command(self, node,
                                              over_write_run_id=self.replace_run_id)
    container_operator = self.create_container_op(working_on=node, command=command)

    for key, _ in parameters.items():
        container_operator.add_env_variable(V1EnvVar(name=defaults.PARAMETER_PREFIX + key,
                                                     value="{{workflow.parameters." + key + "}}"))

    # Disable caching if asked
    if not self.config.enable_caching:
        dsl.get_pipeline_conf().add_op_transformer(disable_cache)


def convert_to_kfp():
    """
    The function that makes the workflow
    """
    dag = _GRAPH
    self = _EXECUTOR

    current_node = dag.start_at
    previous_node = None
    logger.info(f'Rendering job started at {current_node}')

    container_operators = {}
    previous_operator = None
    while True:
        working_on = dag.get_node_by_name(current_node)
        if working_on.is_composite:
            raise NotImplementedError('Composite nodes would not be implemented in Kubeflow, check Argo extension')

        if previous_node == current_node:
            raise Exception('Potentially running in a infinite loop')

        command = utils.get_node_execution_command(self, working_on,
                                                   over_write_run_id=self.replace_run_id)
        container_operator = self.create_container_op(working_on=working_on, command=command)

        if container_operator not in container_operators:
            container_operators[current_node] = container_operator

        if previous_operator:
            container_operator.after(previous_operator)

        previous_node = current_node
        previous_operator = container_operator

        fail_node_name = working_on._get_on_failure_node()

        if fail_node_name:  # Not the Fail node of the graph
            logger.warning('KFP does not support on_failure without significant workarounds! Use Argo extension instead')
            current_node = working_on._get_next_node()
            continue

        if fail_node_name:
            fail_node = dag.get_node_by_name(fail_node_name)
            fail_operator = self.create_container_op(working_on=fail_node)
            if fail_node_name not in container_operators:
                container_operators[fail_node_name] = fail_operator

        if working_on.node_type in ['success', 'fail']:
            break

        current_node = working_on._get_next_node()

    # Disable caching if asked
    if not self.config.enable_caching:
        dsl.get_pipeline_conf().add_op_transformer(disable_cache)



import logging

from magnus import defaults
from magnus.integration import BaseIntegration

logger = logging.getLogger(defaults.NAME)


class BufferedRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """
    executor_type = 'kfp'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'buffered'  # The actual implementation of the service

    def validate(self, **kwargs):
        raise Exception('KFP cannot run work with buffered run log store')


class FileSystemRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """
    executor_type = 'kfp'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'file-system'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "KFP cannot run work with file-system run log store. Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)


class FileSystemCatalog(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """
    executor_type = 'kfp'
    service_type = 'catalog'  # One of secret, catalog, datastore
    service_provider = 'file-system'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "KFP cannot run work with file-system catalog. Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)



from magnus.integration import BaseIntegration


class LocalContainerComputeDBStore(BaseIntegration):
    """
    Integration between local container and DB run log store
    """
    executor_type = 'local-container'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'db'  # The actual implementation of the service

    def configure_for_execution(self, **kwargs):
        # https://stackoverflow.com/questions/24319662/
        connection_string = self.service.config['connection_string']
        if 'localhost' in connection_string:
            # TODO this could be buggy and should be revisited later
            connection_string = connection_string.replace('localhost', 'host.docker.internal')
            self.service.config['connection_string'] = connection_string



import logging
import os
from functools import partial
from pathlib import Path
from typing import Union

import boto3
from magnus import defaults
from pydantic import BaseModel, Extra

logger = logging.getLogger(defaults.NAME)
logging.getLogger('boto3').setLevel(logging.CRITICAL)
logging.getLogger('botocore').setLevel(logging.CRITICAL)
logging.getLogger('s3transfer').setLevel(logging.CRITICAL)


class AWSConfigMixin:
    """
    A Mixin to provide common utilities for AWS services
    """
    class Config(BaseModel, extra=Extra.allow):
        aws_profile: str = ''
        use_credentials: bool = False
        region: str = defaults.AWS_REGION
        aws_credentials_file: str = str(Path.home() / '.aws' / 'credentials')
        aws_access_key_name: str = 'AWS_ACCESS_KEY_ID'
        aws_secret_access_key_name: str = 'AWS_SECRET_ACCESS_KEY'
        aws_session_key_name: str = 'AWS_SESSION_TOKEN'
        role_arn: str = ''
        session_duration_in_seconds: int = 900

    def __init__(self, config=None):
        config = config or {}
        self.config = self.Config(**config)

    def get_aws_profile(self) -> str:
        """
        Return the aws profile as set by the config or None if config is empty

        Returns:
            str: The AWS profile name or ''
        """
        return self.config.aws_profile

    def remove_aws_profile(self):
        """
        Removes the user mentioned AWS profile from the boto session.

        This is useful in cases where we are doing role based auth instead of profile based.
        """
        self.config.aws_profile = ''

    def set_to_use_credentials(self):
        """
        Sets the use_credentials flag in the config to True
        """
        self.config.use_credentials = True

    def get_region_name(self) -> str:
        """
        Return the AWS region as per the config or the default eu-west-1.

        Returns:
            str: The region name as per the config or eu-west-1
        """
        return self.config.region

    def get_aws_credentials_file(self) -> str:
        """
        Return the AWS credentials file if provided in the config or default to $HOME/.aws/credentials

        Returns:
            str: The location of the credentials file
        """
        return self.config.aws_credentials_file

    def get_aws_credentials_from_env(self) -> dict:
        """
        Look for AWS credentials in the environment variables

        Returns:
            dict: AWS credentials as dictionary
        """

        aws_credentials = {}
        try:
            aws_credentials['AWS_ACCESS_KEY_ID'] = os.environ[self.config.aws_access_key_name]
            aws_credentials['AWS_SECRET_ACCESS_KEY'] = os.environ[self.config.aws_secret_access_key_name]
            aws_credentials['AWS_SESSION_TOKEN'] = os.environ.get(self.config.aws_session_key_name, None)
        except KeyError as _e:
            msg = (
                "Expected AWS credentials as part of the environment. Please set AWS_ACCESS_KEY_ID,"
                "AWS_SECRET_ACCESS_KEY as part of the environment"
            )
            raise Exception(msg) from _e

        return aws_credentials

    def get_boto3_session(self):
        """
        Returns a boto3 session

        If we are not using any credentials, we return either a named session or the default session.
        If we are using AWS credentials, we get a session based on the credentials and assume the role.

        Returns:
            [type]: [description]
        """
        profile = self.get_aws_profile()
        region = self.get_region_name()

        # If we are not using credentials from environment variables, return the session
        if not self.config.use_credentials:
            return boto3.session.Session(profile_name=profile, region_name=region)

        # If we are using creds from environment, there are 2 possible scenarios
        # 1). We want a session with just aws access key and secret
        # 2). We want a session with a role_arn which might require session_id set up

        aws_creds = self.get_aws_credentials_from_env()

        partial_sess = partial(boto3.session.Session, aws_access_key_id=aws_creds['AWS_ACCESS_KEY_ID'],
                               aws_secret_access_key=aws_creds['AWS_SECRET_ACCESS_KEY'],
                               region_name=region)

        if not self.config.role_arn:
            logger.info("No Role ARN present in config, assuming a simple session with no STS")
            return partial_sess()  # Enough information to call the partial

        sess = partial_sess(aws_session_token=aws_creds['AWS_SESSION_TOKEN'])

        sts_connection = sess.client('sts')
        assume_role_object = sts_connection.assume_role(
            RoleArn=self.config.role_arn,
            RoleSessionName='temp-session',
            DurationSeconds=self.config.session_duration_in_seconds)
        credentials = assume_role_object['Credentials']

        tmp_access_key = credentials['AccessKeyId']
        tmp_secret_key = credentials['SecretAccessKey']
        tmp_security_token = credentials['SessionToken']

        return boto3.session.Session(
            aws_access_key_id=tmp_access_key,
            aws_secret_access_key=tmp_secret_key, aws_session_token=tmp_security_token,
            region_name=region
        )



import logging

from magnus import defaults
from magnus.integration import BaseIntegration

logger = logging.getLogger(defaults.NAME)


class LocalCompute(BaseIntegration):
    """
    Integration between local and k8's pvc
    """
    executor_type = 'local'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'chunked-k8s-pvc'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "We can't use the local compute k8s pvc store integration."
        )
        raise Exception(msg)


class LocalContainerCompute(BaseIntegration):
    """
    Integration between local-container and k8's pvc
    """
    executor_type = 'local-container'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'chunked-k8s-pvc'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "We can't use the local-container compute k8s pvc store integration."
        )
        raise Exception(msg)


class KfpCompute(BaseIntegration):
    """
    Integration between kfp and k8's pvc
    """
    executor_type = 'kfp'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'chunked-k8s-pvc'  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        self.executor.persistent_volumes["run-log-store"] = (
            self.service.config.persistent_volume_name, self.service.config.mount_path)


class ArgoCompute(BaseIntegration):
    """
    Integration between argo and k8's pvc
    """
    executor_type = 'argo'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'chunked-k8s-pvc'  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        self.executor.persistent_volumes["run-log-store"] = (
            self.service.config.persistent_volume_name, self.service.config.mount_path)



import logging
from pathlib import Path

from magnus import defaults
from magnus.integration import BaseIntegration

logger = logging.getLogger(defaults.NAME)


class LocalContainerCompute(BaseIntegration):
    """
    Integration between local container and S3 run log store
    """
    executor_type = 'local-container'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'chunked-s3'  # The actual implementation of the service

    def validate(self, **kwargs):
        pass

    def configure_for_traversal(self, **kwargs):
        write_to = self.service.get_aws_credentials_file()
        self.executor.volumes[str(Path(write_to).resolve())] = {
            'bind': '/root/.aws/credentials',
            'mode': 'ro'
        }



import datetime
import json
import logging
from collections import OrderedDict
from string import Template as str_template

from magnus import defaults, exceptions
from magnus.datastore import BaseRunLogStore, BranchLog, RunLog, StepLog

logger = logging.getLogger(defaults.NAME)
logger.info('Loading DB datastore extension')


try:
    import sqlalchemy
    from sqlalchemy import Column, DateTime, Integer, Sequence, Text
    from sqlalchemy.orm import declarative_base, sessionmaker

    Base = declarative_base()

    class DBLog(Base):
        """
        Base table for storing run logs in database.

        In this model, we fragment the run log into logical units that are concurrent safe.
        """
        __tablename__ = 'db_log'
        pk = Column(Integer, Sequence('id_seq'), primary_key=True)
        run_id = Column(Text)
        attribute_key = Column(Text)  # run_log, step_internal_name, parameter_key etc
        attribute_type = Column(Text)  # RunLog, Step, Branch, Parameter
        attribute_value = Column(Text)  # The JSON string
        created_at = Column(DateTime, default=datetime.datetime.utcnow)

        def __repr__(self):
            return f'DBLog for {self.run_id} and {self.attribute_key}'

except ImportError as _e:
    logger.exception('Unable to import SQLalchemy, is it installed?')
    msg = (
        "SQLAlchemy is required for this extension. Please install it"
    )
    raise Exception(msg) from _e


def create_tables(connection_string):
    """
    from magnus.datastore_extensions import db
    connection_string = 'postgresql://localhost/'
    db.create_tables(connection_string)
    """
    engine = sqlalchemy.create_engine(connection_string)
    Base.metadata.create_all(engine)


class DBStore(BaseRunLogStore):
    """
    Using SQL alchemy to interface for a database as a Run Log store.
    All concurrently accessible attributes are stored in independent rows

    The expected columns:
        run_id : str,
        attribute_key: str,  # step name, branch name, parameter name etc
        attribute_type: str, # step, branch, parameter, run_log
        attribute_value: json, # string representation of JSON
        created_at: UTC time now,  #Needed for ordering the Run log

    Example config:
    run_log:
      type: db
      config:
        connection_string: The connection string to use in SQLAlchemy. Secret placeholders are fine.
    """
    service_name = 'db'

    def __init__(self, config):
        super().__init__(config)
        if 'connection_string' not in self.config:
            raise Exception('Run Log stores of DB should have a connection string')

        self.engine = None  # Follows a singleton pattern
        self.session = None

        # Attribute Types
        self.RUNLOG_ATTRIBUTE_TYPE = 'RunLog'  # pylint: disable=c0103
        self.PARAMETER_ATTRIBUTE_TYPE = 'Parameter'  # pylint: disable=c0103
        self.STEPLOG_ATTRIBUTE_TYPE = 'StepLog'  # pylint: disable=c0103
        self.BRANCHLOG_ATTRIBUTE_TYPE = 'BranchLog'  # pylint: disable=c0103

    @property
    def connection_string(self) -> str:
        """
        Returns the connection string as provided by the config

        Raises:
            Exception: If connection_string is not provided in the config

        Returns:
            str: The connection string specified in the config
        """
        return self.config['connection_string']

    def _make_db_engine(self):
        """
        Creates a DB engine and session object once per execution. Singleton pattern
        """
        if not self.engine:
            from magnus.pipeline import \
                global_executor  # pylint: disable=C0415

            secrets = global_executor.secrets_handler.get()  # Returns all secrets as dictionary
            connection_string = str_template(self.connection_string).safe_substitute(**secrets)

            self.engine = sqlalchemy.create_engine(connection_string, pool_pre_ping=True)
            self.session = sessionmaker(bind=self.engine)

    def write_to_db(self, run_id: str, attribute_key: str, attribute_type: str, attribute_value: str):
        """
        Write a part of the Run log to the database against a specified run_id

        Args:
            run_id (str): The run id to update the run log
            attribute_key (str): run_log for RunLog, the step log internal name for steps, parameter key for parameter
            attribute_type (str): One of RunLog, Parameter, StepLog, BranchLog
            attribute_value (str): The value to put in the database
        """
        self._make_db_engine()

        with self.session() as session:
            record = DBLog(run_id=run_id, attribute_key=attribute_key,
                           attribute_type=attribute_type, attribute_value=attribute_value)
            session.add(record)
            session.commit()

    def update_db(self, run_id: str, attribute_type: str, attribute_key: str, attribute_value: str, upsert=True):
        """
        Update the Database with a part of run log.

        If upsert is true, we create the object instead of updating.

        Args:
            run_id (str): The run id to update the run log
            attribute_key (str): run_log for RunLog, the step log internal name for steps, parameter key for parameter
            attribute_type (str): One of RunLog, Parameter, StepLog, BranchLog
            attribute_value (str): The value to put in the database
            upsert (bool, optional): Create if not present. Defaults to True.
        """
        self._make_db_engine()

        with self.session() as session:
            records = session.query(DBLog).filter(DBLog.run_id == run_id). \
                filter(DBLog.attribute_type == attribute_type).\
                filter(DBLog.attribute_key == attribute_key).all()

            for record in records:
                record.attribute_value = attribute_value

            if upsert and len(records) == 0:
                # We insert otherwise
                record = DBLog(run_id=run_id, attribute_key=attribute_key,
                               attribute_type=attribute_type, attribute_value=attribute_value)
                session.add(record)

            session.commit()

    def get_from_db(self, run_id: str, attribute_type: str, attribute_key=None):
        """
        Gets an attribute from the db.
        The attribute could be a RunLog, StepLog, BranchLog or Parameter

        Args:
            run_id (str): The Run Id to retrieve the attribute
            attribute_type (str): The attribute to retrieve from the DB for the run_id
            attribute_key (str, optional): Conditional filtering of attribute. Defaults to None.

        Returns:
            object: All the records from the DB which match the run_id and attribute
        """
        self._make_db_engine()

        with self.session() as session:
            query = session.query(DBLog).order_by(DBLog.created_at).filter(DBLog.run_id == run_id). \
                filter(DBLog.attribute_type == attribute_type)

            if attribute_key:
                query = query.filter(DBLog.attribute_key == attribute_key)
            records = query.all()

            return records

    def _get_parent_branch(self, name: str) -> str:  # pylint: disable=R0201
        """
        Returns the name of the parent branch.
        If the step is part of main dag, return None.

        Args:
            name (str): The name of the step.

        Returns:
            str: The name of the branch containing the step.
        """
        dot_path = name.split('.')

        if len(dot_path) == 1:
            return None
        # Ignore the step name
        return '.'.join(dot_path[:-1])

    def _get_parent_step(self, name: str) -> str:  # pylint: disable=R0201
        """
        Returns the step containing the step, useful when we have steps within a branch.
        Returns None, if the step belongs to parent dag.

        Args:
            name (str): The name of the step to find the parent step it belongs to.

        Returns:
            str: The parent step the step belongs to, None if the step belongs to parent dag.
        """
        dot_path = name.split('.')

        if len(dot_path) == 1:
            return None
        # Ignore the branch.step_name
        return '.'.join(dot_path[:-2])

    def prepare_full_run_log(self, run_log: RunLog) -> RunLog:
        """
        Populates the run log with the branches and steps.
        In database mode, we fragment the run log into individual Run Logs and this method populates the
        full run log by querying the underlying tables.

        Args:
            run_log (RunLog): The partial run log containing empty step logs
        """
        run_id = run_log.run_id
        run_log.parameters = self.get_parameters(run_id=run_id)

        all_steps = self.get_from_db(run_id=run_id, attribute_type=self.STEPLOG_ATTRIBUTE_TYPE)
        all_branches = self.get_from_db(run_id=run_id, attribute_type=self.BRANCHLOG_ATTRIBUTE_TYPE)

        ordered_steps = OrderedDict()
        for step in all_steps:
            json_str = json.loads(step.attribute_value)
            ordered_steps[step.attribute_key] = StepLog(**json_str)

        ordered_branches = OrderedDict()
        for branch in all_branches:
            json_str = json.loads(branch.attribute_value)
            ordered_branches[branch.attribute_key] = BranchLog(**json_str)

        current_branch = None
        for step_internal_name in ordered_steps:
            current_branch = self._get_parent_branch(step_internal_name)
            step_to_add_branch = self._get_parent_step(step_internal_name)

            if not current_branch:
                current_branch = run_log
            else:
                current_branch = ordered_branches[current_branch]
                step_to_add_branch = ordered_steps[step_to_add_branch]
                step_to_add_branch.branches[current_branch.internal_name] = current_branch

            current_branch.steps[step_internal_name] = ordered_steps[step_internal_name]

    def create_run_log(self, run_id, **kwargs):
        """
        Creates a Run Log object by using the config

        Logically the method should do the following:
            * Creates a Run log
            * Adds it to the db
            * Return the log
        """
        logger.info(f'{self.service_name} Creating a Run Log for : {run_id}')
        run_log = RunLog(run_id=run_id, status=defaults.CREATED)
        self.write_to_db(run_id=run_id,
                         attribute_key='run_log',
                         attribute_type=self.RUNLOG_ATTRIBUTE_TYPE,
                         attribute_value=json.dumps(run_log.dict(), ensure_ascii=True))  # pylint: disable=no-member

        return run_log

    def get_run_log_by_id(self, run_id, full=True, **kwargs):
        """
        Retrieves a Run log from the database using the config and the run_id

        Args:
            run_id (str): The run_id of the run

        Logically the method should:
            * Returns the run_log defined by id from the data store defined by the config

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        try:
            logger.info(f'{self.service_name} Getting a Run Log for : {run_id}')
            records = self.get_from_db(run_id=run_id,
                                       attribute_key='run_log',
                                       attribute_type=self.RUNLOG_ATTRIBUTE_TYPE)
            if not records:
                raise exceptions.RunLogNotFoundError(run_id)

            json_str = json.loads(records[0].attribute_value)

            run_log = RunLog(**json_str)

            if full:
                self.prepare_full_run_log(run_log)
            return run_log
        except:
            raise exceptions.RunLogNotFoundError(run_id)

    def put_run_log(self, run_log, **kwargs):
        """
        Puts the Run Log in the database as defined by the config

        Args:
            run_log (RunLog): The Run log of the run

        Logically the method should:
            Puts the run_log into the database
        """
        run_id = run_log.run_id
        logger.info(f'{self.service_name} Putting the Run Log for : {run_id}')
        self.update_db(run_id=run_id,
                       attribute_key='run_log',
                       attribute_type=self.RUNLOG_ATTRIBUTE_TYPE,
                       attribute_value=json.dumps(run_log.dict(), ensure_ascii=True))  # pylint: disable=no-member

    def get_parameters(self, run_id, **kwargs):
        """
        Get the parameters from the Run log defined by the run_id

        Args:
            run_id (str): The run_id of the run

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Return the parameters as identified in the run_log

        Returns:
            dict: A dictionary of the run_log parameters
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        logger.info(f'{self.service_name} Getting parameters for : {run_id}')
        records = self.get_from_db(run_id=run_id, attribute_type=self.PARAMETER_ATTRIBUTE_TYPE)

        parameters = {}
        for record in records:
            parameters[record.attribute_key] = json.loads(record.attribute_value)

        return parameters

    def set_parameters(self, run_id, parameters, **kwargs):
        """
        Update the parameters of the Run log with the new parameters

        This method would over-write the parameters, if the parameter exists in the run log already

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Update the parameters of the run_log
            * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            run_id (str): The run_id of the run
            parameters (dict): The parameters to update in the run log
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        logger.info(f'{self.service_name} Getting parameters for : {run_id}')
        for key, value in parameters.items():
            self.update_db(run_id=run_id,
                           attribute_type=self.PARAMETER_ATTRIBUTE_TYPE,
                           attribute_key=key,
                           attribute_value=json.dumps(value)
                           )

    def get_step_log(self, internal_name, run_id, **kwargs):
        """
        Get a step log from the datastore for run_id and the internal naming of the step log

        The internal naming of the step log is a dot path convention.

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Identify the step location by decoding the internal naming
            * Return the step log

        Args:
            internal_name (str): The internal name of the step log
            run_id (str): The run_id of the run

        Returns:
            StepLog: The step log object for the step defined by the internal naming and run_id

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            StepLogNotFoundError: If the step log for internal_name is not found in the datastore for run_id
        """
        logger.info(f'{self.service_name} Getting the step log: {internal_name} of {run_id}')

        records = self.get_from_db(run_id=run_id, attribute_type=self.STEPLOG_ATTRIBUTE_TYPE,
                                   attribute_key=internal_name)
        if len(records) > 1:
            raise Exception(f'{self.service_name} Getting too many step logs for {internal_name}')

        if not records:
            raise exceptions.StepLogNotFoundError(run_id=run_id, name=internal_name)

        json_str = json.loads(records[0].attribute_value)
        return StepLog(**json_str)

    def add_step_log(self, step_log, run_id, **kwargs):
        """
        Add the step log in the run log as identifed by the run_id in the datastore

        The method should:
             * Call get_run_log_by_id(run_id) to retrieve the run_log
             * Identify the branch to add the step by decoding the step_logs internal name
             * Add the step log to the identified branch log
             * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            step_log (StepLog): The Step log to add to the database
            run_id (str): The run id of the run

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            BranchLogNotFoundError: If the branch of the step log for internal_name is not found in the datastore
                                    for run_id
        """
        logger.info(f'{self.service_name} Adding the step log to DB: {step_log.name}')
        self.update_db(run_id=run_id,
                       attribute_key=step_log.internal_name,
                       attribute_type=self.STEPLOG_ATTRIBUTE_TYPE,
                       attribute_value=json.dumps(step_log.dict(), ensure_ascii=True))  # pylint: disable=no-member

    def get_branch_log(self, internal_branch_name, run_id, **kwargs):
        # Should get the run_log
        # Should search for the branch log with the name
        logger.info(f'{self.service_name} Getting the Branch log: {internal_branch_name} of {run_id}')

        if not internal_branch_name:
            return self.get_run_log_by_id(run_id=run_id)

        records = self.get_from_db(run_id=run_id, attribute_type=self.BRANCHLOG_ATTRIBUTE_TYPE,
                                   attribute_key=internal_branch_name)
        if len(records) > 1:
            raise Exception(f'{self.service_name} Getting too many branch logs for {internal_branch_name}')

        if not records:
            raise exceptions.BranchLogNotFoundError(run_id=run_id, name=internal_branch_name)

        json_str = json.loads(records[0].attribute_value)
        return BranchLog(**json_str)

    def add_branch_log(self, branch_log, run_id, **kwargs):
        # This could in some occasions be the RunLog
        # Get the run log
        # Get the branch and step containing the branch
        # Add the branch to the step
        # Write the run_log
        if type(branch_log) == RunLog:
            self.put_run_log(run_log=branch_log)
            return
        logger.info(f'{self.service_name} Adding the Branch log to DB: {branch_log.internal_name}')
        self.update_db(run_id=run_id,
                       attribute_key=branch_log.internal_name,
                       attribute_type=self.BRANCHLOG_ATTRIBUTE_TYPE,
                       attribute_value=json.dumps(branch_log.dict(), ensure_ascii=True))



import fnmatch
import json
import logging
import tempfile
import time
from pathlib import Path
from string import Template
from typing import Any, List, Optional, Union

import botocore
from magnus import defaults
from magnus.datastore import ChunkedFileSystemRunLogStore

logger = logging.getLogger(defaults.NAME)


try:
    from magnus_extensions.aws import AWSConfigMixin
except ImportError as _e:  # pragma: no cover
    msg = (
        "AWS Dependencies are not installed!!"
    )
    raise Exception(msg) from _e


class ChunkedS3Store(ChunkedFileSystemRunLogStore, AWSConfigMixin):
    """
    S3 implementation of Chunked Run Log store

    Example config:
    run_log_store:
      type: s3
      config:
        s3_bucket: The S3 bucket to use
        prefix: Any bucket prefix that you want to attach

    """
    service_name = 'chunked-s3'
    LogTypes = ChunkedFileSystemRunLogStore.LogTypes

    class Config(AWSConfigMixin.Config):
        s3_bucket: str
        prefix: str = ''

    @property
    def s3_bucket(self) -> str:
        """
        Returns the S3 bucket name from the config

        Returns:
            str: The name of the s3 bucket as per the config
        """
        return self.config.s3_bucket

    @property
    def prefix(self):
        """
        Return the prefix if present in the config. Or an empty string
        """
        return Path(self.config.prefix.strip('/'))

    def prefix_with_run_id(self, run_id: str) -> Path:
        return self.prefix / run_id

    def get_matches(self, run_id: str, name: str, multiple_allowed: bool = False) -> Optional[Union[List[Path], Path]]:
        """
        Get contents of files matching the pattern name*

        Args:
            run_id (str): The run id
            name (str): The suffix of the file name to check in the run log store.
        """
        sub_name = Template(name).safe_substitute({"creation_time": ""})
        search_name = str(f'{self.prefix_with_run_id(run_id=run_id)}/{sub_name}*')

        boto3_session = self.get_boto3_session()
        s3_client = boto3_session.client('s3')
        page_iter = s3_client.get_paginator('list_objects_v2').paginate(
            Bucket=self.s3_bucket, Prefix=str(self.prefix_with_run_id(run_id=run_id)))

        matches = []
        for page in page_iter:
            try:
                matches += fnmatch.filter([file['Key'] for file in page['Contents']], search_name)
            except KeyError:
                logger.warning(f"Did not find any objects matching the pattern")
                return None

        if not matches:
            return None

        if not multiple_allowed:
            if len(matches) > 1:
                msg = (
                    f"Multiple matches found for {name} while multiple is not allowed"
                )
                raise Exception(msg)
            return Path(matches[0])

        return [Path(match) for match in matches]

    def _store(self, run_id: str, contents: dict, name: Path):
        """
        Store the contents against the name in the folder.

        Args:
            run_id (str): The run id
            contents (dict): The dict to store
            name (str): The name to store as
        """
        boto3_session = self.get_boto3_session()
        s3_client = boto3_session.client('s3')

        with tempfile.NamedTemporaryFile() as temp_file:
            with open(temp_file.name, 'w', encoding='utf-8') as fw:
                json.dump(contents, fw, ensure_ascii=True, indent=4)  # pylint: disable=no-member

            object_key = self.prefix_with_run_id(run_id=run_id) / name.name
            try:
                s3_client.upload_file(temp_file.name, self.s3_bucket, str(object_key))
            except botocore.exceptions.ClientError as _e:
                if _e.response['Error']['Code'] == "404":
                    logger.exception(f'Upload failed to {self.s3_bucket}')
                raise Exception(_e) from _e

    def _retrieve(self, name: Path) -> dict:
        """
        Does the job of retrieving from the folder.

        Args:
            name (str): the name of the file to retrieve

        Returns:
            dict: The contents
        """
        boto3_session = self.get_boto3_session()
        s3_client = boto3_session.client('s3')

        with tempfile.NamedTemporaryFile() as temp_file:
            key = str(name)

            logger.info(f'Trying to download {key} from {self.s3_bucket}')
            try:
                s3_client.download_file(self.s3_bucket, key, temp_file.name)
                json_str = json.load(open(temp_file.name, 'rb'))
                return json_str
            except botocore.exceptions.ClientError as _e:
                if _e.response['Error']['Code'] == "404":
                    logger.exception(f'{key} does not exist in {self.s3_bucket}')
                    raise _e

    def store(self, run_id: str, log_type: LogTypes, contents: dict, name: str = ''):
        """Store a SINGLE log type in the file system

        Args:
            run_id (str): The run id to store against
            log_type (LogTypes): The type of log to store
            contents (dict): The dict of contents to store
            name (str, optional): The name against the contents have to be stored. Defaults to ''.
        """

        naming_pattern = self.naming_pattern(log_type=log_type, name=name)
        match = self.get_matches(run_id=run_id, name=naming_pattern, multiple_allowed=False)
        # The boolean multiple allowed confuses mypy a lot!
        name_to_give: Path = None  # type: ignore
        if match:
            existing_contents = self._retrieve(name=match)  # type: ignore
            contents = dict(existing_contents, **contents)
            name_to_give = match  # type: ignore
        else:
            _name = Template(naming_pattern).safe_substitute({"creation_time": str(int(time.time_ns()))})
            name_to_give = self.prefix_with_run_id(run_id=run_id) / _name

        self._store(run_id=run_id, contents=contents, name=name_to_give)

    def retrieve(self, run_id: str, log_type: LogTypes, name: str = '', multiple_allowed=False) -> Any:
        """
        Retrieve the model given a log_type and a name.
        Use multiple_allowed to control if you are expecting multiple of them.
        eg: There could be multiple of Parameters- but only one of StepLog-stepname

        The reasoning for name to be defaulted to empty string:
            Its actually conditionally empty. For RunLog and Parameter it is empty.
            For StepLog and BranchLog it should be provided.

        Args:
            run_id (str): The run id
            log_type (LogTypes): One of RunLog, Parameter, StepLog, BranchLog
            name (str, optional): The name to match. Defaults to ''.
            multiple_allowed (bool, optional): Are multiple allowed. Defaults to False.

        Raises:
            FileNotFoundError: If there is no match found

        Returns:
            Any: One of StepLog, BranchLog, Parameter or RunLog
        """
        if not name and log_type not in [self.LogTypes.RUN_LOG, self.LogTypes.PARAMETER]:
            raise Exception(f"Name is required during retrieval for {log_type}")

        naming_pattern = self.naming_pattern(log_type=log_type, name=name)
        matches = self.get_matches(run_id=run_id, name=naming_pattern, multiple_allowed=multiple_allowed)

        if matches:
            if not multiple_allowed:
                contents = self._retrieve(name=matches)  # type: ignore
                model = self.ModelTypes[log_type.name].value
                return model(**contents)

            models = []
            for match in matches:  # type: ignore
                contents = self._retrieve(name=match)
                model = self.ModelTypes[log_type.name].value
                models.append(model(**contents))
            return models

        raise FileNotFoundError()




import logging
import shlex
from typing import Dict, List

from magnus import defaults, integration, utils
from magnus.executor import BaseExecutor
from magnus.graph import Graph
from magnus.nodes import BaseNode
from ruamel.yaml import YAML

logger = logging.getLogger(defaults.NAME)

try:
    from kubernetes import client
    from kubernetes import config as k8s_config
    from kubernetes.client import (V1EnvVar, V1EnvVarSource,
                                   V1PersistentVolumeClaimVolumeSource,
                                   V1SecretKeySelector)
except ImportError as _e:
    msg = (
        "Kubernetes Dependencies have not been installed!!"
    )
    raise Exception(msg) from _e


class K8sJobExecutor(BaseExecutor):

    service_name = "k8s-job"

    class Config(BaseExecutor.Config):
        config_path: str
        docker_image: str
        namespace: str = "default"
        cpu_limit: str = "250m"
        memory_limit: str = "1G"
        gpu_limit: int = 0
        gpu_vendor: str = "nvidia.com/gpu"
        cpu_request: str = ""
        memory_request: str = ""
        active_deadline_seconds: int = 60 * 60 * 2  # 2 hours
        ttl_seconds_after_finished: int = 60  #  1 minute
        image_pull_policy: str = "Always"
        secrets_from_k8s: dict = {}  # EnvVar=SecretName:Key
        persistent_volumes: dict = {}  # volume-name:mount_path
        labels: Dict[str, str] = {}

    def __init__(self, config: dict = None):
        super().__init__(config)
        self.persistent_volumes = {}

        for i, (claim, mount_path) in enumerate(self.config.persistent_volumes.items()):
            self.persistent_volumes[f"executor-{i}"] = (claim, mount_path)

    def prepare_for_graph_execution(self):
        """
        This method would be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we need not do anything.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_traversal(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_traversal(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_traversal(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_traversal(self, self.experiment_tracker)

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_execution(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_execution(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_execution(self, self.secrets_handler)

        integration.validate(self, self.experiment_tracker)
        integration.configure_for_execution(self, self.experiment_tracker)

        self._set_up_run_log(exists_ok=True)

    @property
    def _client(self):
        k8s_config.load_kube_config(config_file=self.config.config_path)
        return client

    def execute_job(self, node: BaseNode):
        command = utils.get_job_execution_command(self, node)
        logger.info(f'Triggering a kubernetes job with : {command}')

        self.config.labels['job_name'] = self.run_id

        k8s_batch = self._client.BatchV1Api()

        cpu_limit = self.config.cpu_limit
        memory_limit = self.config.memory_limit

        cpu_request = self.config.cpu_request or cpu_limit
        memory_request = self.config.memory_request or memory_limit

        gpu_limit = str(self.config.gpu_limit)  # Should be something like nvidia -etc

        limits = {"cpu": cpu_limit, "memory": memory_limit, self.config.gpu_vendor: gpu_limit}
        requests = {"cpu": cpu_request, "memory": memory_request}
        resources = {"limits": limits, "requests": requests}

        environment_variables = []
        for secret_env, k8_secret in self.config.secrets_from_k8s.items():
            try:
                secret_name, key = k8_secret.split(':')
            except Exception as _e:
                msg = (
                    "K8's secret should be of format EnvVar=SecretName:Key"
                )
                raise Exception(msg) from _e
            secret_as_env = V1EnvVar(name=secret_env, value_from=V1EnvVarSource(
                secret_key_ref=V1SecretKeySelector(name=secret_name, key=key)))
            environment_variables.append(secret_as_env)

        overridden_params = utils.get_user_set_parameters()
        # The parameters present in the environment override the parameters present in the parameters file
        # The values are coerced to be strings, hopefully they will be fine on the other side.
        for k, v in overridden_params.items():
            environment_variables.append(V1EnvVar(name=defaults.PARAMETER_PREFIX+k, value=str(v)))

        pod_volumes = []
        volume_mounts = []
        for claim_name, (claim, mount_path) in self.persistent_volumes.items():
            pod_volumes.append(self._client.V1Volume(
                name=claim_name, persistent_volume_claim=V1PersistentVolumeClaimVolumeSource(claim_name=claim)))
            volume_mounts.append(self._client.V1VolumeMount(name=claim_name, mount_path=mount_path))

        base_container = self._client.V1Container(
            name=self.run_id,
            image=self.config.docker_image,
            command=shlex.split(command),
            resources=resources,
            env=environment_variables,
            image_pull_policy="Always",
            volume_mounts=volume_mounts or None
        )

        pod_spec = self._client.V1PodSpec(volumes=pod_volumes or None,
                                          restart_policy='Never',
                                          containers=[base_container])

        pod_template = self._client.V1PodTemplateSpec(metadata=client.V1ObjectMeta(
            labels=self.config.labels, annotations={"sidecar.istio.io/inject": "false"}), spec=pod_spec)

        job_spec = client.V1JobSpec(template=pod_template, backoff_limit=2,
                                    ttl_seconds_after_finished=self.config.ttl_seconds_after_finished)
        job_spec.active_deadline_seconds = self.config.active_deadline_seconds

        job = client.V1Job(
            api_version='batch/v1',
            kind='Job',
            metadata=client.V1ObjectMeta(name=self.run_id),
            spec=job_spec)

        logger.debug(f'Submitting kubernetes job: {job}')

        try:
            response = k8s_batch.create_namespaced_job(
                body=job, namespace=self.config.namespace, _preload_content=False, pretty=True)
            print(f'Kubernetes job {self.run_id} created')
            logger.debug(f'Kubernetes job response: {response}')
        except Exception as e:
            logger.exception(e)
            raise

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self.run_log_store.add_step_log(step_log, self.run_id)

        super()._execute_node(node, map_variable=map_variable, **kwargs)

        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
        if step_log.status == defaults.FAIL:
            raise Exception(f'Step {node.name} failed')

    def execute_graph(self, dag: Graph, map_variable: dict = None, **kwargs):
        msg = (
            "This executor is not supported to execute any graphs but only jobs (functions or notebooks)"
        )
        raise NotImplementedError(msg)

    def send_return_code(self, stage='traversal'):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        if stage != 'traversal':  # traversal does no actual execution, so return code is pointless
            run_id = self.run_id

            run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
            if run_log.status == defaults.FAIL:
                raise Exception('Pipeline execution failed')



import json
import logging
from functools import lru_cache

from magnus import defaults, exceptions
from magnus.secrets import BaseSecrets

logger = logging.getLogger(defaults.NAME)

try:
    from magnus_extensions.aws import AWSConfigMixin
except ImportError as _e:  # pragma: no cover
    msg = (
        "AWS Dependencies are not installed!!"
    )
    raise Exception(msg) from _e


class AWSSecretsManager(BaseSecrets, AWSConfigMixin):
    """
    Implementation of secrets manager of AWS

    Example config:
    secrets:
      type: 'aws-secrets-manager'
      config:
        secret_arn: The secret ARN to retrieve the secrets from.
        region: The aws region to use
        aws_profile: The aws profile to use. The executor which uses the secrets manager can also control
                this behavior by the integration pattern.
    """
    service_name = 'aws-secrets-manager'

    class Config(AWSConfigMixin.Config):
        secret_arn: str
        secret_version_id: str = ""

    def __init__(self, config: dict, **kwargs):
        super().__init__(config, **kwargs)
        self.secrets = {}

    @property
    def secret_arn(self) -> str:
        """Returns the secret arn stored in the config

        Returns:
            str: The arn of the secret
        """
        return self.config.secret_arn

    @property
    def secret_version(self) -> str:
        """
        Returns the version ID if provided in the config

        Returns:
            str: The version ID as specified in the config or None
        """
        return self.config.secret_version_id

    @lru_cache(maxsize=2)
    def load_secrets(self):
        """
        One time load of all the secrets
        """
        boto_session = self.get_boto3_session()

        client = boto_session.client('secretsmanager')
        if self.secret_version:
            response = client.get_secret_value(SecretId=self.secret_arn, VersionId=self.secret_version)
        else:
            response = client.get_secret_value(SecretId=self.secret_arn)
        self.secrets = json.loads(response['SecretString'])

    def get(self, name: str = None, **kwargs):
        """
        Return the secret by name.
        If no name is give, return all the secrets.

        Args:
            name (str): The name of the secret to return.
        """
        self.load_secrets()
        if not name:
            return self.secrets

        if name in self.secrets:
            return self.secrets[name]

        secrets_location = self.get_secret_arn()
        raise exceptions.SecretNotFoundError(secret_name=name, secret_setting=secrets_location)
